from django.conf import settings
from django.conf.urls import url
from django.core.exceptions import ImproperlyConfigured, ViewDoesNotExist
from django.http import HttpResponse, StreamingHttpResponse, Http404, HttpResponseRedirect, HttpResponsePermanentRedirect
from django.template import TemplateDoesNotExist, TemplateSyntaxError

from .exceptions import InternalRedirectException, RedirectException, DMPViewDoesNotExist
from .signals import dmp_signal_pre_process_request, dmp_signal_post_process_request, dmp_signal_internal_redirect_exception, dmp_signal_redirect_exception
from .template import TemplateViewFunction
from .util import get_dmp_instance, get_dmp_app_configs, log, DMP_OPTIONS
from .util import DMP_VIEW_ERROR, DMP_VIEW_FUNCTION, DMP_VIEW_CLASS_METHOD, DMP_VIEW_TEMPLATE

import os, os.path, re, mimetypes, sys, logging, pkgutil
from urllib.parse import unquote
from importlib import import_module



##############################################################
###   The front controller of all views on the site.
###   urls.py routes everything through this method.

def route_request(request, *args, **kwargs):
    '''
    The main router for all calls coming in to the system.  Patterns in urls.py should call this function.
    '''
    # wrap to enable the InternalRedirectExceptions to loop around
    response = None
    while True:
        # an outer try that catches the redirect exceptions
        try:
            # output the variables so the programmer can debug where this is routing
            if log.isEnabledFor(logging.INFO):
                log.info('processing: app={}, page={}, module={}, func={}, urlparams={}'.format(request.dmp_router_app, request.dmp_router_page, request.dmp_router_module, request.dmp_router_function, request.urlparams))

            # ensure we have a dmp_router_callback variable on request
            if getattr(request, 'dmp_router_callback', None) is None:
                raise ImproperlyConfigured("Variable request.dmp_router_callback does not exist (check MIDDLEWARE for `django_mako_plus.RequestInitMiddleware`).")

            # if we had a view not found, raise a 404
            if isinstance(request.dmp_router_callback, ViewDoesNotExist) or request.dmp_router_callback._dmp_view_type is DMP_VIEW_ERROR:
                log.error(str(request.dmp_router_callback))
                raise Http404

            # send the pre-signal
            if DMP_OPTIONS.get('SIGNALS', False):
                for receiver, ret_response in dmp_signal_pre_process_request.send(sender=sys.modules[__name__], request=request):
                    if isinstance(ret_response, (HttpResponse, StreamingHttpResponse)):
                        return ret_response

            # log the view
            if log.isEnabledFor(logging.INFO):
                if request.dmp_router_callback._dmp_view_type is DMP_VIEW_CLASS_METHOD:
                    log.info('calling class-based view function {}.{}.{}'.format(request.dmp_router_module, request.dmp_router_class, request.dmp_router_function))
                elif request.dmp_router_callback._dmp_view_type is DMP_VIEW_TEMPLATE:
                    log.info('view function {}.{} not found; rendering template {}'.format(request.dmp_router_module, request.dmp_router_function, request.dmp_router_fallback))
                else: # assume a view function
                    log.info('calling view function {}.{}'.format(request.dmp_router_module, request.dmp_router_function))

            # call view function with any args and any remaining kwargs
            response = request.dmp_router_callback(request, *args, **kwargs)

            # send the post-signal
            if DMP_OPTIONS.get('SIGNALS', False):
                for receiver, ret_response in dmp_signal_post_process_request.send(sender=sys.modules[__name__], request=request, response=response):
                    if ret_response != None:
                        response = ret_response # sets it to the last non-None in the signal receiver chain

            # if we didn't get a correct response back, send a 404
            if not isinstance(response, (HttpResponse, StreamingHttpResponse)):
                if request.dmp_router_class == None:
                    log.error('view function {}.{} failed to return an HttpResponse (or the post-signal overwrote it).  Returning Http404.'.format(request.dmp_router_module, request.dmp_router_function))
                else:
                    log.error('class-based view function {}.{}.{} failed to return an HttpResponse (or the post-signal overwrote it).  Returning Http404.'.format(request.dmp_router_module, request.dmp_router_class, request.dmp_router_function))
                raise Http404

            # return the response
            return response

        except InternalRedirectException as ivr:
            # send the signal
            if DMP_OPTIONS.get('SIGNALS', False):
                dmp_signal_internal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=ivr)
            # resolve to a function
            request.dmp_router_module = ivr.redirect_module
            request.dmp_router_function = ivr.redirect_function
            try:
                module_obj = import_module(request.dmp_router_module)
                request.dmp_router_callback = getattr(module_obj, request.dmp_router_function, None)
                if request.dmp_router_callback == None:
                    request.dmp_router_callback = DMPViewDoesNotExist('Module {} found successfully during internal redirect, but view function {} is not defined in the module.'.format(request.dmp_router_module, request.dmp_router_function))
            except ImportError:
                request.dmp_router_callback = DMPViewDoesNotExist('View function {}.{} not found during internal redirect.'.format(request.dmp_router_module, request.dmp_router_function))
            # do the internal redirect
            log.info('received an InternalViewRedirect to {} -> {}'.format(request.dmp_router_module, request.dmp_router_function))

        except RedirectException as e: # redirect to another page
            if request.dmp_router_class == None:
                log.info('view function {}.{} redirected processing to {}'.format(request.dmp_router_module, request.dmp_router_function, e.redirect_to))
            else:
                log.info('class-based view function {}.{}.{} redirected processing to {}'.format(request.dmp_router_module, request.dmp_router_class, request.dmp_router_function, e.redirect_to))
            # send the signal
            if DMP_OPTIONS.get('SIGNALS', False):
                dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
            # send the browser the redirect command
            return e.get_response(request)

    # the code should never get here
    raise Exception("Django-Mako-Plus error: The route_request() function should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")





###############################################################
###   A decorator that signals that a function is callable
###   by DMP.  This prevents "just any function" from being
###   called by the router above.
###

def view_function(f):
    '''
    A decorator to signify which view functions are "callable" by web browsers.

    Any endpoint function, such as process_request, must be decorated to be callable:

        @view_function
        function process_request(request):
            ...

    The @view_function decorator must be the first one listed if other decorators are present.
    Note that class-based views don't need to be decorated because we allow anything that extends Django's View class.
    '''
    # This decorator is more of an annotation on functions than a decorator.
    # Rather than the usual inner function pattern, I'm simply setting a flag on the
    # function to signify that it is callable.  The router checks this flag
    # before calling the function.
    f._dmp_view_type = DMP_VIEW_FUNCTION
    return f






from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

import django_mako_plus

import os, os.path
from importlib import import_module


class Command(BaseCommand):
    args = '<name>'
    help = 'Creates a new Django-Mako-Plus app.'
    can_import_settings = True
    missing_args_message = "You must provide an application name."


    def add_arguments(self, parser):
        # required argument for the app name
        parser.add_argument('appname', type=str, help='the name of the new app')


    def handle(self, *args, **options):
        self.stdout.write("Error:")
        self.stdout.write('')
        self.stdout.write("As of v1.7, Django-Mako-Plus apps are creted with the standard Django command.")
        self.stdout.write("Please see https://github.com/doconix/django-mako-plus#create-a-dmp-style-app")
        self.stdout.write('')
        self.stdout.write("\t\t\t\t-- The Management")




from django.apps import apps, AppConfig
from django.conf import settings
from django.conf.urls import url
from django.core.exceptions import ImproperlyConfigured
from django.template import TemplateDoesNotExist, engines, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.context import _builtin_context_processors
from django.utils.module_loading import import_string
from django.views.generic import View

from .exceptions import InternalRedirectException, RedirectException, DMPViewDoesNotExist
from .signals import dmp_signal_pre_render_template, dmp_signal_post_render_template, dmp_signal_redirect_exception
from .template import MakoTemplateLoader, MakoTemplateAdapter, TemplateViewFunction
from .util import get_dmp_instance, get_dmp_app_configs, log, DMP_OPTIONS, DMP_INSTANCE_KEY
from .util import DMP_VIEW_FUNCTION, DMP_VIEW_CLASS_METHOD

from mako.template import Template

from copy import deepcopy
from inspect import isclass
from importlib import import_module
from importlib.util import find_spec
import os, os.path, sys, itertools, threading, collections

try:
    # python 3.4+
    from importlib.util import find_spec
except ImportError:
    # python <= 3.3
    from importlib import find_loader as find_spec


#########################################################
###   The main engine

class MakoTemplates(BaseEngine):
    '''
    The primary Mako interface that plugs into the Django templating system.
    This is referenced in settings.py -> TEMPLATES.
    '''
    def __init__(self, params):
        '''Constructor'''
        # pop the settings.py options into the DMP options
        try:
            DMP_OPTIONS.update(params.pop('OPTIONS'))
        except KeyError:
            raise ImproperlyConfigured('The Django Mako Plus template OPTIONS were not set up correctly in settings.py.  Please ensure the OPTIONS is set in the template setup.')

        # cache our instance in the util module for get_dmp_instance
        # this is a bit of a hack, but it makes calling utility methods possible and efficient
        DMP_OPTIONS[DMP_INSTANCE_KEY] = self
        self.template_loaders = {}
        self.dmp_enabled_apps = set()
        self.view_cache = {}
        self.rlock = threading.RLock()

        # super constructor
        super(MakoTemplates, self).__init__(params)

        # set up the context processors
        context_processors = []
        for processor in itertools.chain(_builtin_context_processors, DMP_OPTIONS.get('CONTEXT_PROCESSORS', [])):
            context_processors.append(import_string(processor))
        self.template_context_processors = tuple(context_processors)

        # now that our engine has loaded, initialize a few parts of it
        # should we minify JS AND CSS FILES?
        DMP_OPTIONS['RUNTIME_JSMIN'] = False
        DMP_OPTIONS['RUNTIME_CSSMIN'] = False
        if DMP_OPTIONS.get('MINIFY_JS_CSS', False) and not settings.DEBUG:
            try:
                from rjsmin import jsmin
            except ImportError:
                raise ImproperlyConfigured('MINIFY_JS_CSS = True in the Django Mako Plus settings, but the "rjsmin" package does not seem to be loaded.')
            try:
                from rcssmin import cssmin
            except ImportError:
                raise ImproperlyConfigured('MINIFY_JS_CSS = True in the Django Mako Plus settings, but the "rcssmin" package does not seem to be loaded.')
            DMP_OPTIONS['RUNTIME_JSMIN'] = jsmin
            DMP_OPTIONS['RUNTIME_CSSMIN'] = cssmin

        # should we compile SASS files?
        DMP_OPTIONS['RUNTIME_SCSS_ENABLED'] = False
        SCSS_BINARY = DMP_OPTIONS.get('SCSS_BINARY', None)
        if isinstance(SCSS_BINARY, str):  # for backwards compatability
            log.warning('Future warning: the settings.py variable SCSS_BINARY should be a list of arguments, not a string.')
            DMP_OPTIONS['RUNTIME_SCSS_ARGUMENTS'] = SCSS_BINARY.split(' ')
            DMP_OPTIONS['RUNTIME_SCSS_ENABLED'] = True
        elif isinstance(SCSS_BINARY, (list, tuple)):
            DMP_OPTIONS['RUNTIME_SCSS_ARGUMENTS'] = SCSS_BINARY
            DMP_OPTIONS['RUNTIME_SCSS_ENABLED'] = True
        elif not SCSS_BINARY:
            DMP_OPTIONS['RUNTIME_SCSS_ARGUMENTS'] = None
        else:
            raise ImproperlyConfigured('The SCSS_BINARY option in Django Mako Plus settings must be a list of arguments.  See the DMP documentation.')

        # set up the template imports (we always include django_mako_plus in the links)
        DMP_OPTIONS['ALL_TEMPLATE_IMPORTS'] = list(DMP_OPTIONS.get('DEFAULT_TEMPLATE_IMPORTS', []))
        DMP_OPTIONS['ALL_TEMPLATE_IMPORTS'].append('import django_mako_plus')

        # add a template renderer for each DMP-enabled app
        for app_config in get_dmp_app_configs():
            self.register_app(app_config)


    def register_app(self, app):
        '''
        Registers an app as a "DMP-enabled" app.  Registering creates a cached
        template renderer to make processing faster and adds the dmp_render()
        and dmp_render_to_string() methods to the app.  When Django starts,
        this method is called automatically for any app with
        DJANGO_MAKO_PLUS = True in its __init__.py file.

        app: The name of the app or an AppConfig instance.
        '''
        # ensure we have an AppConfig object (recurse to all DMP apps if None)
        if isinstance(app, str):
            app = apps.get_app_config(app)

        # add to the set of DMP-enabled apps
        self.dmp_enabled_apps.add(app.name)

        # set up the template, script, and style renderers
        # these create and cache just by accessing them
        self.get_template_loader(app, 'templates', create=True)
        self.get_template_loader(app, 'scripts', create=True)
        self.get_template_loader(app, 'styles', create=True)

        # add the shortcut functions (only to the main templates, we don't do to scripts or styles
        # because people generally don't call those directly).  This is a monkey patch, but it is
        # an incredibly useful one because it makes calling app-specific rendering functions much
        # easier.
        #
        # Django's shortcut to return an *HttpResponse* is render(), and its template method to render a *string* is also render().
        # Good job on naming there, folks.  That's going to confuse everyone.  But I'm matching it to be consistent despite the potential confusion.
        with self.rlock:
            app.module.dmp_render_to_string = _render_to_string(self, app.name)
            app.module.dmp_render = _render(self, app.name)


    def is_dmp_app(self, app):
        '''
        Returns True if the given app is a DMP-enabled app.  The app parameter can
        be either the name of the app or an AppConfig object.
        '''
        if isinstance(app, AppConfig):
            app = app.name
        return app in self.dmp_enabled_apps


    def from_string(self, template_code):
        '''
        Compiles a template from the given string.
        This is one of the required methods of Django template engines.
        '''
        mako_template = Template(template_code, imports=DMP_OPTIONS.get('DEFAULT_TEMPLATE_IMPORTS'), input_encoding=DMP_OPTIONS.get('DEFAULT_TEMPLATE_ENCODING', 'utf-8'))
        return MakoTemplateAdapter(mako_template)


    def get_view_function(self, app_name, module_name, function_name, fallback_template_name):
        '''
        Returns the view function for the given app_name + module_name + function_name.
        If not found, it returns the view function for the fallback_template_name within app_name.
        If still not found, raises a ViewDoesNotExist exception.

        This method uses a cache for a significant speedup.  The cache use is thread safe.
        It primarily used by the DMP router.
        '''
        # get the function (or error) from the cache
        cache_key = ( app_name, module_name, function_name )
        try:
            return self.view_cache[cache_key]
        except KeyError:
            # acquire the lock and try the cache again (another thread might have put it in cache while we are waiting for lock)
            with self.rlock:
                try:
                    return self.view_cache[cache_key]
                except KeyError:  # really not in the cache, so let's go get it
                    # check if the module exists. i'm using find_spec instead of import_module
                    # to know if it exists because import_module fails on things like syntax error,
                    # and the programmer should see these.
                    func_obj = None
                    if find_spec(module_name) is not None:
                        module_obj = import_module(module_name)
                        func_obj = getattr(module_obj, function_name, None)
                        if func_obj == None:
                            func_obj = DMPViewDoesNotExist('Module {} found successfully, but view function {} is not defined in the module.'.format(module_name, function_name))

                    # check for not found, no @view_function, View subclass
                    if func_obj == None:
                        # try to load the template directly
                        try:
                            func_obj = TemplateViewFunction(app_name, fallback_template_name)
                            func_obj.get_template()  # check whether the template exists
                        except TemplateDoesNotExist as e:
                            func_obj = DMPViewDoesNotExist('View function {}.{} not found, and fallback template {} not found.'.format(module_name, function_name, fallback_template_name))

                    elif isclass(func_obj) and issubclass(func_obj, View):
                        # this Django method wraps the view class with a function, so now we can treat it like a regular dmp view function
                        func_obj = func_obj.as_view()
                        func_obj._dmp_view_type = DMP_VIEW_CLASS_METHOD  # have to monkey patch in this case because we don't control the as_view() method of View

                    # the function must be DMP_VIEW_* (see util.py)
                    if getattr(func_obj, '_dmp_view_type', None) == None:
                        func_obj = DMPViewDoesNotExist('View function {}.{} found successfully, but it must be decorated with @view_function.  Note that if you have multiple decorators on a function, the @view_function decorator must be listed first.'.format(module_name, function_name))

                    # cache the result (if production)
                    if not settings.DEBUG:
                        self.view_cache[cache_key] = func_obj

                    # return the obj!
                    return func_obj

        # the code should never be able to get here
        raise Exception("Django-Mako-Plus error: The get_view_function() function should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")


    def get_template(self, template_name):
        '''
        Retrieves a template object.
        This is one of the required methods of Django template engines.

        Because DMP templates are always app-specific (Django only searches
        a global set of directories), the template_name MUST be in the format:
        "app_name/template.html".  DMP splits the template_name string on the
        slash to get the app name and template name.

        Only forward slashes are supported.  Use / even on Windows to specify
        the "app_name/template.html".
        '''
        parts = template_name.split('/', 1)
        if len(parts) < 2:
            raise TemplateDoesNotExist('Invalid template_name format for a DMP template.  This method requires that the template name be in app_name/template.html format (separated by slash).')
        return self.get_template_loader(parts[0]).get_template('/'.join(parts[1:]))


    def get_template_loader(self, app, subdir='templates', create=False):
        '''
        Returns a template loader object for the given app name in the given subdir.
        For example, get_template_loader('homepage', 'styles') will return
        a loader for the styles/ directory in the homepage app.

        The app parameter can be either an app name or an AppConfig instance.
        The subdir parameter is normally 'templates', 'scripts', or 'styles',
        but it can be any subdirectory name of the given app.

        Normally, you should not have to call this method.  Django automatically
        generates two shortcut functions for every DMP-registered app (apps with
        DJANGO_MAKO_PLUS = True), and these shortcut functions are the preferred
        way to render templates.

        This method is useful when you want a custom template loader to a directory
        that does not conform to the app_dir/templates/* pattern.

        If the loader is not found in the DMP cache, one of two things occur:
          1. If create=True, it is created automatically and returned.  This overrides
             the need to set DJANGO_MAKO_PLUS=True in the app's __init__.py file.
          2. If create=False, a TemplateDoesNotExist is raised.  This is the default
             behavior.
        '''
        # ensure we have an AppConfig
        if not isinstance(app, AppConfig):
            app = apps.get_app_config(app)

        # get the loader with the path of this app+subdir
        path = os.path.join(app.path, subdir)

        # if create=False, the loader must already exist in the cache
        if not create and path not in self.template_loaders:
            raise TemplateDoesNotExist("%s has not been registered as a DMP app.  Did you forget to include the DJANGO_MAKO_PLUS=True line in your app's __init__.py?" % app.name)

        # return the template by path
        return self.get_template_loader_for_path(path, use_cache=True)


    def get_template_loader_for_path(self, path, use_cache=True):
        '''
        Returns a template loader object for the given directory path.
        For example, get_template_loader('/var/mytemplates/') will return
        a loader for that specific directory.

        Normally, you should not have to call this method.  Django automatically
        generates two shortcut functions for every DMP-registered app (apps with
        DJANGO_MAKO_PLUS = True), and these shortcut functions are the preferred
        way to render templates.

        This method is useful when you want a custom template loader for a specific
        directory that may be outside your project directory or that is otherwise
        not contained in a normal Django app.  If the directory is inside an app,
        call get_template_loader() instead.

        Unless use_cache=False, this method caches template loaders in the DMP
        cache for later use.
        '''
        # get from the cache if we are able
        if use_cache:
            try:
                return self.template_loaders[path]
            except KeyError:
                pass  # not there, so we'll create

        # create the loader
        loader = MakoTemplateLoader(path, None)

        # cache if we are allowed
        if use_cache:
            self.template_loaders[path] = loader

        # return
        return loader



############################################################################
###  Monkey-patch functions to enable render() and render_to_string()
###  within the app scope of *each* DMP-enabled app.


def _render_to_string(dmp_instance, app_name):
    # I'm doing this inner function for late lookups (get_template_loader), just in case new template loader objects are added after creation.
    def wrapper(request, template, context=None, def_name=None, subdir='templates'):
        '''
        A shortcut to render a template.  This is one of the primary functions in the DMP framework.
        This method is added to the app space of each DMP-enabled app at load time.

            @request      The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                          file will be ignored but the template will otherwise render fine.
            @template     The template file path to render.  This is relative to the app_path/controller_TEMPLATES_DIR/ directory.
                          For example, to render app_path/templates/page1, set template="page1.html", assuming you have
                          set up the variables as described in the documentation above.
            @context      A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                          or a Django Context object.
            @def_name     Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                          For example, def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within the template.
            @subdir       The sub-folder within the app where the template resides.  Templates are normally placed
                          in project/appname/templates/, which is the default value.

        Returns the rendered template as a unicode string.

        Examples of use from within appname/views/someview.py:

            from django_mako_plus import view_function
            from .. import render

            @view_function
            def process_request(request):
                # render an HTML template in appname/templates/sometemplate.html to a string
                html = render_to_string(request, 'sometemplate.html', { 'var1': 'value' })
                # (do something with the html)

            @view_function
            def another_func(request):
                # render some dynamic JS in appname/scripts/sometemplate.js (which has embedded Mako code)
                html = render_to_string(request, 'sometemplate.js', { 'var1': 'value' }, subdir="scripts")
                # (do something with the html)

            @view_function
            def yet_another(request):
                # render a single <block name="toolbar"> instead of the entire template in appname/templates/sometemplate.html to a string
                html = render_to_string(request, 'sometemplate.html', { 'var1': 'value' }, def_name='toolbar')
                # (do something with the html)

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        template_loader = dmp_instance.get_template_loader(app_name, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render')(context=context, request=request, def_name=def_name)

    # outer function return
    return wrapper


def _render(dmp_instance, app_name):
    # I'm doing this inner function for late lookups (get_template_loader), just in case new template loader objects are added after creation.
    def wrapper(request, template, context=None, def_name=None, subdir='templates', content_type=None, status=None, charset=None):
        '''
        A shortcut to render a template.  This is one of the primary functions in the DMP framework.
        This method is added to the app space of each DMP-enabled app at load time.

            @request      The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                          file will be ignored but the template will otherwise render fine.
            @template     The template file path to render.  This is relative to the app_path/controller_TEMPLATES_DIR/ directory.
                          For example, to render app_path/templates/page1, set template="page1.html", assuming you have
                          set up the variables as described in the documentation above.
            @context      A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                          or a Django Context object.
            @def_name     Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                          For example, def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within the template.
            @subdir       The sub-folder within the app where the template resides.  Templates are normally placed
                          in project/appname/templates/, which is the default value.
            @content_type The MIME type of the response.  Defaults to settings.DEFAULT_CONTENT_TYPE (usually 'text/html').
            @status       The HTTP response status code.  Defaults to 200 (OK).
            @charset      The charset to encode the processed template string (the output) with.  Defaults to settings.DEFAULT_CHARSET (usually 'utf-8').

        Returns a Django HttpResponse containing the rendered template.

        Examples of use from within appname/views/someview.py:

            from django_mako_plus import view_function
            from .. import render

            @view_function
            def process_request(request):
                # return an HTML template in appname/templates/sometemplate.html
                return render_to_string(request, 'sometemplate.html', { 'var1': 'value' })

            @view_function
            def another_func(request):
                # return some dynamic JS in appname/scripts/sometemplate.js (which has embedded Mako code)
                return render_to_string(request, 'sometemplate.js', { 'var1': 'value' }, subdir="scripts")

            @view_function
            def yet_another(request):
                # return a single <block name="toolbar"> instead of the entire template in appname/templates/sometemplate.html
                return render_to_string(request, 'sometemplate.html', { 'var1': 'value' }, def_name='toolbar')

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        template_loader = dmp_instance.get_template_loader(app_name, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render_to_response')(context=context, request=request, def_name=def_name, content_type=content_type, status=status, charset=charset)

    # outer function return
    return wrapper




from django.core.exceptions import ImproperlyConfigured
from django.views.generic import View
from urllib.parse import unquote

# try to import MiddlewareMixIn (Django 1.10+)
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # create a dummy MiddlewareMixin if older Django
    MiddlewareMixin = object

from .router import route_request
from .util import URLParamList, get_dmp_instance, DMP_OPTIONS, log
from .util import DMP_VIEW_CLASS_METHOD

import logging

##########################################################
###   Middleware the prepares the request for
###   use with the controller.

EMPTY_URLPARAMS = URLParamList()


class RequestInitMiddleware(MiddlewareMixin):
    '''
    Adds several fields to the request that our controller needs.
    Projects can customize the variables with view middleware that runs after this class.

    This class must be included in settings.py -> MIDDLEWARE.
    '''

    def process_request(self, request):
        '''
        Adds stubs for the DMP custom variables to the request object.

        We do this early in the middleware process because process_view() below gets called
        *after* URL resolution.  If an exception occurs before that (a 404, a middleware exception),
        the variables won't be put on the request object.  Therefore, we put stubs on it at the
        earliest possible place so they exist even if we don't get to process_view().
        '''
        request.dmp_router_app = None
        request.dmp_router_page = None
        request.dmp_router_function = None
        request.dmp_router_module = None
        request.dmp_router_class = None
        request.urlparams = EMPTY_URLPARAMS
        request.dmp_router_callback = None


    def process_view(self, request, view_func, view_args, view_kwargs):
        '''
        Adds the following to the request object:

            request.dmp_router_app        The Django application (such as "homepage"), as a string.
            request.dmp_router_page       The view module (such as "index" for index.py), as a string.
            request.dmp_router_function   The function within the view module to be called (usually "process_request"), as a string.
            request.dmp_router_module     The module path in Python terms (such as homepage.views.index), as a string.
            request.dmp_router_class      This is set to None in this method, but route_request() fills it in, as a string, if a class-based view.
            request.urlparams             A list of the remaining url parts, as a list of strings.  See the tutorial for more information on this parameter.
            request.dmp_router_callback   The view callable (function, method, etc.) to be called by the router.

        Named parameters in the url pattern determines the values of these variables.
        See django_mako_plus/urls.py for the DMP standard patterns.  For example, one pattern
        specifies the router app, page, function, and urlparams: /app/page.function/urlparams

            ^(?P<dmp_router_app>[_a-zA-Z0-9\-]+)/(?P<dmp_router_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_router_function>[_a-zA-Z0-9\.\-]+)/?(?P<urlparams>.*)$

        As view middleware, this function runs just before the router.route_request is called.
        '''
        # print debug messages to help with urls.py regex issues
        if log.isEnabledFor(logging.DEBUG):
            if view_func is not route_request:
                log.debug("DMP is not handling this request because the matched urls.py regex did not resolve to route_request. The request.dmp_* variables will not be set.")
            else:
                for param in ( 'dmp_router_app', 'dmp_router_page', 'dmp_router_function', 'urlparams' ):
                    if param not in view_kwargs:
                        log.debug('The matched urls.py regex did not contain the named parameter "{}"; using the default value from settings.py.'.format(param))

        # if urls.py didn't resolve to DMP, we don't need to set any of these variables
        if view_func is not route_request:
            return

        # add the variables to the request
        request.dmp_router_app = view_kwargs.pop('dmp_router_app', None) or DMP_OPTIONS.get('DEFAULT_APP', 'homepage')
        request.dmp_router_page = view_kwargs.pop('dmp_router_page', None) or DMP_OPTIONS.get('DEFAULT_PAGE', 'index')
        request.dmp_router_function = view_kwargs.pop('dmp_router_function', None)
        if request.dmp_router_function:
            request.dmp_router_fallback = '{}.{}.html'.format(request.dmp_router_page, request.dmp_router_function)
        else:
            request.dmp_router_fallback = '{}.html'.format(request.dmp_router_page)
            request.dmp_router_function = 'process_request'

        # period and dash cannot be in python names, but we allow dash in app and page and (dash or period) in the function
        # these change into underscores
        request.dmp_router_app = request.dmp_router_app.replace('-', '_')
        request.dmp_router_page = request.dmp_router_page.replace('-', '_')
        request.dmp_router_function = request.dmp_router_function.replace('.', '_').replace('-', '_')

        # add the full module path to the request
        request.dmp_router_module = '.'.join([ request.dmp_router_app, 'views', request.dmp_router_page ])

        # add the url parameters to the request
        # note that I'm not using unquote_plus because the + switches to a space *after* the question mark (in the regular parameters)
        # in the normal url, spaces should be quoted with %20.  Thanks Rosie for the tip.
        request.urlparams = URLParamList(( unquote(s) for s in view_kwargs.pop('urlparams', '').split('/') ))

        # get the function object - the return of get_view_function might be a function, a class-based view, or a template
        # get_view_function does some magic to make all of these act like a regular view function
        request.dmp_router_callback = get_dmp_instance().get_view_function(request.dmp_router_app, request.dmp_router_module, request.dmp_router_function, request.dmp_router_fallback)

        # adjust the variable values if a class
        if request.dmp_router_callback._dmp_view_type == DMP_VIEW_CLASS_METHOD:
            request.dmp_router_class = request.dmp_router_function
            request.dmp_router_function = request.method.lower()
        # debugging
        # print('request.dmp_router_app        ', request.dmp_router_app        )
        # print('request.dmp_router_page       ', request.dmp_router_page       )
        # print('request.dmp_router_function   ', request.dmp_router_function   )
        # print('request.dmp_router_module     ', request.dmp_router_module     )
        # print('request.dmp_router_class      ', request.dmp_router_class      )
        # print('request.urlparams             ', request.urlparams             )
        # print('request.dmp_router_callback   ', request.dmp_router_callback   )



################################################################
###   A set of request processors that add variables to the
###   context (parameters) when templates are rendered.
###

from django.conf import settings as conf_settings


def settings(request):
    '''Adds the settings dictionary to the request'''
    return { 'settings': conf_settings }



from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from django_mako_plus.util import get_dmp_instance, get_dmp_app_configs, DMP_OPTIONS

from optparse import make_option
import os, os.path, shutil, fnmatch
from importlib import import_module


# import minification if requested
JSMIN = False
CSSMIN = False
if DMP_OPTIONS.get('MINIFY_JS_CSS', False):
    try:
        from rjsmin import jsmin
        JSMIN = True
    except ImportError:
        raise CommandError('The Django Mako Plus option "MINIFY_JS_CSS" is True in settings.py, but the "rjsmin" module does not seem to be installed. Do you need to "pip install" it?')
    try:
        from rcssmin import cssmin
        CSSMIN = True
    except ImportError:
        raise CommandError('The Django Mako Plus option "MINIFY_JS_CSS" is True in settings.py, but the "rcssmin" module does not seem to be installed. Do you need to "pip install" it?')



class Command(BaseCommand):
    args = ''
    help = 'Collects static files, such as media, scripts, and styles, to a common directory root. This is done to prepare for deployment.'
    can_import_settings = True
    
    
    def add_arguments(self, parser):
        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing files in the directory when necessary.'
        )
        parser.add_argument(
            '--ignore',
            action='append',
            dest='ignore_files',
            help='Ignore the given file/directory.  Unix-style wildcards are acceptable, such as "*.txt".  This option can be specified more than once.'
        )


    def handle(self, *args, **options):
        # save the options for later
        self.options = options

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting. Aborting app creation.')

        # get the destination directory, and ensure it doesn't already exist
        # if dest_root starts with a /, it is an absolute directory
        # if dest_root doesn't start with a /, it goes relative to the BASE_DIR
        try:
            dest_root = os.path.join(os.path.abspath(settings.BASE_DIR), settings.STATIC_ROOT)
            if os.path.exists(dest_root) and not self.options['overwrite']:
                raise CommandError('The destination directory for static files (%s) already exists. Please delete it or run this command with the --overwrite option.' % dest_root)
        except AttributeError:
            raise CommandError('Your settings.py file is missing the STATIC_ROOT setting. Exiting without collecting the static files.')

        # create the directory - we assume it either doesn't exist, or the --overwrite is specified
        if not os.path.isdir(dest_root):
            os.makedirs(dest_root)

        # go through the DMP apps and collect the static files
        for config in get_dmp_app_configs():
            self.copy_dir(config.path, os.path.abspath(os.path.join(dest_root, config.name)))



    def ignore_file(self, fname):
        '''Returns whether the given filename should be ignored, based on the --ignore options sent into the command'''
        if self.options['ignore_files']:
            for pattern in self.options['ignore_files']:
                if fnmatch.fnmatch(fname, pattern):
                    return True
        return False


    def copy_dir(self, source, dest, level=0):
        '''Copies the static files from one directory to another.  If this command is run, we assume the user wants to overwrite any existing files.'''
        # ensure the destination exists
        if not os.path.exists(dest):
            os.mkdir(dest)
        # go through the files in this directory
        for fname in os.listdir(source):
            source_path = os.path.join(source, fname)
            dest_path = os.path.join(dest, fname)
            ext = os.path.splitext(fname)[1].lower()

            ###  EXPLICIT IGNORE  ###
            if self.ignore_file(fname):
                pass

            ###  DIRECTORIES  ###
            # ignore these directories
            elif os.path.isdir(source_path) and fname in ( 'migrations', 'templates', 'views', DMP_OPTIONS.get('TEMPLATES_CACHE_DIR'), '__pycache__' ):
                pass

            # if a directory, create it in the destination and recurse
            elif os.path.isdir(source_path):
                if not os.path.exists(dest_path):
                    os.mkdir(dest_path)
                elif not os.path.isdir(dest_path):  # could be a file or link
                    os.unlink(dest_path)
                    os.mkdir(dest_path)
                self.copy_dir(source_path, dest_path, level+1)

            ###   FILES   ###
            # we don't do any regular files at the top level
            elif level == 0:
                pass

            # ignore these files
            elif fname in ( '__init__.py', ):
                pass

            # ignore these extensions
            elif ext in ( '.cssm', '.jsm' ):
                pass

            # if a regular Javscript file, minify it
            elif ext == '.js' and DMP_OPTIONS.get('MINIFY_JS_CSS', False) and JSMIN:
                with open(source_path) as fin:
                    with open(dest_path, 'w') as fout:
                        fout.write(jsmin(fin.read()))

            elif ext == '.css' and DMP_OPTIONS.get('MINIFY_JS_CSS', False) and CSSMIN:
                with open(source_path) as fin:
                    with open(dest_path, 'w') as fout:
                        fout.write(cssmin(fin.read()))

            # if we get here, it's a binary file like an image, movie, pdf, etc.
            else:
                shutil.copy2(source_path, dest_path)
